###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

from typing import Callable, Any
from cmflib import cli
from cmflib.cmf_exception_handling import CmfResponse

def exception_handler_decorator(
    target_function: Callable[..., Any]
) -> Callable[..., Any]:
    """
    Decorator to handle CmfResponse exceptions.
    """
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return target_function(*args, **kwargs).handle()
        except CmfResponse as e:
            return e.handle()

    return wrapper


@exception_handler_decorator
def _metadata_push(
    pipeline_name: str, file_name: str, execution_uuid: str, tensorboard_path: str
) -> str:
    """ Pushes metadata file to CMF-server.
    Args:
        pipeline_name: Name of the pipeline.
        file_name: Specify input metadata file name.
        execution_uuid: Optional execution UUID.
        tensorboard_path: Path to tensorboard logs.
    Returns:
        Output from the metadata push command.
    """
    cli_args = [
            "metadata",
            "push",
            "-p",
            pipeline_name,
            "-f",
            file_name,
        ]
    
    # Only append the execution_uuid and tensorboard path flag if a real value was supplied
    if execution_uuid:
        cli_args.extend(["-e", execution_uuid])
    if tensorboard_path:
       cli_args.extend(["-t", tensorboard_path])

    
    cli_args = cli.parse_args(cli_args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _metadata_pull(pipeline_name: str, file_name: str, execution_uuid: str) -> str:
    """ Pulls metadata file from CMF-server. 
     Args: 
        pipeline_name: Name of the pipeline. 
        file_name: Specify output metadata file name.
        execution_uuid: Optional execution UUID. 
     Returns: 
        Output from the metadata pull command. 
     """
    # Here, We are not extending the cli_args list directly with execution_uuid 
    # bcaz in our code[cmf/commands/metadata/pull.py], we are assinging execution_uuid as a none 
    # if it is not present 
    cli_args = cli.parse_args(
        [
            "metadata",
            "pull",
            "-p",
            pipeline_name,
            "-f",
            file_name,
            "-e",
            execution_uuid,
        ]
    )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _metadata_export(pipeline_name: str, json_file_name: str, file_name: str) -> str:
    """ Export local metadata's metadata in json format to a json file. 
     Args: 
        pipeline_name: Name of the pipeline. 
        json_file_name: File path of json file. 
        file_name: Specify input metadata file name. 
     Returns: 
        Output from the metadata export command. 
     """
    cli_args = [
            "metadata",
            "export",
            "-p",
            pipeline_name,
            "-f",
            file_name,
        ]
    #  Only append the json_file_name flag if a real value was supplied
    if json_file_name:
       cli_args.extend(["-j", json_file_name])

    cli_args = cli.parse_args(cli_args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _artifact_push(pipeline_name: str, file_name: str, jobs: str) -> str:
    """ Pushes artifacts to the initialized repository.
    Args: 
       pipeline_name: Name of the pipeline. 
       filepath: Path to store the artifact. 
       jobs: Number of jobs to use for pushing artifacts.
    Returns:
        Output from the artifact push command.
    """
    cli_args = cli.parse_args(
        [
            "artifact",
            "push",
            "-p",
            pipeline_name,
            "-f",
            file_name,
            "-j",
            jobs,
        ]
    )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _artifact_pull(pipeline_name: str, file_name: str, artifact_name: str) -> str:
    """ Pulls artifacts from the initialized repository.
    Args:
        pipeline_name: Name of the pipeline.
        file_name: Specify input metadata file name.
        artifact_name: Name of the artifact. 
    Returns:
        Output from the artifact pull command.
    """
    cli_args = [
            "artifact",
            "pull",
            "-p",
            pipeline_name,
            "-f",
            file_name,
        ]
    # Only append the optional artifact name flag if a real value was supplied
    if artifact_name:
         cli_args.extend(["-a", artifact_name])
         
    cli_args = cli.parse_args(cli_args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _cmf_init_show() -> str:
    """ Initializes and shows details of the CMF command. 
    Returns: 
       Output from the init show command. 
    """ 
    cli_args = cli.parse_args(["init", "show"])
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _init_local(
    path: str,
    git_remote_url: str,
    cmf_server_url: str,
    neo4j_user: str,
    neo4j_password: str,
    neo4j_uri: str,
) -> str:
    """Initialize local repository"""
    args = [
            "init",
            "local",
            "--path",
            path,
            "--git-remote-url",
            git_remote_url,
            "--cmf-server-url",
            cmf_server_url,
        ]

    # only append neo4j args if they are provided
    if neo4j_user and neo4j_password and neo4j_uri:
        args.extend(
            [
                "--neo4j-user",
                neo4j_user,
                "--neo4j-password",
                neo4j_password,
                "--neo4j-uri",
                neo4j_uri,
            ]
        )
    
    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _init_minioS3(
    url: str,
    endpoint_url: str,
    access_key_id: str,
    secret_key: str,
    git_remote_url: str,
    cmf_server_url: str,
    neo4j_user: str,
    neo4j_password: str,
    neo4j_uri: str,
) -> str:
    """Initialize minioS3 repository"""
    args = [
            "init",
            "minioS3",
            "--url",
            url,
            "--endpoint-url",
            endpoint_url,
            "--access-key-id",
            access_key_id,
            "--secret-key",
            secret_key,
            "--git-remote-url",
            git_remote_url,
            "--cmf-server-url",
            cmf_server_url,
        ]
    # only append neo4j args if they are provided
    if neo4j_user and neo4j_password and neo4j_uri:
        args.extend(
            [
                "--neo4j-user",
                neo4j_user,
                "--neo4j-password",
                neo4j_password,
                "--neo4j-uri",
                neo4j_uri,
            ]
        )
    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg
    

@exception_handler_decorator
def _init_amazonS3(
    url: str,
    access_key_id: str,
    secret_key: str,
    session_token: str,
    git_remote_url: str,
    cmf_server_url: str,
    neo4j_user: str,
    neo4j_password: str,
    neo4j_uri: str,
) -> str:
    """Initialize amazonS3 repository"""
    args = [
            "init",
            "amazonS3",
            "--url",
            url,
            "--access-key-id",
            access_key_id,
            "--secret-key",
            secret_key,
            "--session-token",
            session_token,
            "--git-remote-url",
            git_remote_url,
            "--cmf-server-url",
            cmf_server_url,
        ]

    # only append neo4j args if they are provided
    if neo4j_user and neo4j_password and neo4j_uri:
        args.extend(
            [
                "--neo4j-user",
                neo4j_user,
                "--neo4j-password",
                neo4j_password,
                "--neo4j-uri",
                neo4j_uri,
            ]
        )
    
    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _init_sshremote(
    path: str,
    user: str,
    port: str,
    password: str,
    git_remote_url: str,
    cmf_server_url: str,
    neo4j_user: str,
    neo4j_password: str,
    neo4j_uri: str,
) -> str:
    """Initialize sshremote repository"""
    args = [
            "init",
            "sshremote",
            "--path",
            path,
            "--user",
            user,
            "--port",
            port,
            "--password",
            password,
            "--git-remote-url",
            git_remote_url,
            "--cmf-server-url",
            cmf_server_url,
        ]
    # only append neo4j args if they are provided
    if neo4j_user and neo4j_password and neo4j_uri:
        args.extend(
            [
                "--neo4j-user",
                    neo4j_user,
                    "--neo4j-password",
                    neo4j_password,
                    "--neo4j-uri",
                    neo4j_uri,
                ]
            )

    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _init_osdfremote(
    path: str,
    cache: str,
    key_id: str,
    key_path: str,
    key_issuer: str,
    git_remote_url: str,
    cmf_server_url: str,
    neo4j_user: str,
    neo4j_password: str,
    neo4j_uri: str,
) -> str:
    """Initialize osdfremote repository"""
    args = [
            "init",
            "osdf",
            "--path",
            path,
            "--cache",
            cache,
            "--key-id",
            key_id,
            "--key-path",
            key_path,
            "--key-issuer",
            key_issuer,
            "--git-remote-url",
            git_remote_url,
            "--cmf-server-url",
            cmf_server_url,
    ]
    # only append neo4j args if they are provided
    if neo4j_user and neo4j_password and neo4j_uri:
        args.extend(
            [
                "--neo4j-user",
                    neo4j_user,
                    "--neo4j-password",
                    neo4j_password,
                    "--neo4j-uri",
                    neo4j_uri,
                ]
            )
    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg
  

@exception_handler_decorator
def _artifact_list(pipeline_name: str, file_name: str, artifact_name: str) -> str:
    """Displays artifacts from the input metadata file with a few properties
    in a 7-column table, limited to 20 records per page.
    Args:
       pipeline_name: Name of the pipeline.
       file_name: Specify input metadata file name.
       artifact_name: Artifacts for particular artifact name.
    Returns:
       Output from the artifact list command.
    """
    cli_args = [
            "artifact",
            "list",
            "-p",
            pipeline_name,
            "-f",
            file_name,
        ]
    # Only append the optional artifact name flag if a real value was supplied
    if artifact_name:
        cli_args.extend(["-a", artifact_name])

    cli_args = cli.parse_args(cli_args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _pipeline_list(file_name: str) -> str:
    """ Display a list of pipeline name(s) from the available input metadata file.
    Args:
        file_name: Specify input metadata file name. 
    Returns:
        Output from the pipeline list command.
    """
    cli_args = cli.parse_args(["pipeline", "list", "-f", file_name])
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    return msg


@exception_handler_decorator
def _execution_list(pipeline_name: str, file_name: str, execution_uuid: str) -> str:
    """Displays executions from the input metadata file with a few properties
    in a 7-column table, limited to 20 records per page.
    Args:
       pipeline_name: Name of the pipeline.
       file_name: Specify input metadata file name.
       execution_uuid: Specify the execution uuid to retrieve execution.
    Returns:
       Output from the execution list command.
    """
    args = [
        "execution",
        "list",
        "-p",
        pipeline_name,
        "-f",
        file_name,
    ]
    # Only append the optional execution uuid flag if a real value was supplied
    if execution_uuid:
        args.extend(["-e", execution_uuid])

    cli_args = cli.parse_args(args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _repo_push(
    pipeline_name: str,
    file_name: str,
    tensorboard_path: str,
    execution_uuid: str,
    jobs: str,
) -> str:
    """Push artifacts, metadata files, and source code to the user's artifact
    repository, cmf-server, and git respectively.
    Args:
       pipeline_name: Name of the pipeline.
       file_name: Specify input metadata file name.
       execution_uuid: Specify execution uuid.
       tensorboard_path: Path to tensorboard logs.
       jobs: Number of jobs to use for pushing artifacts.
    Returns:
       Output from the repo push command.
    """
    cli_args = [
            "repo",
            "push",
            "-p",
            pipeline_name,
            "-f",
            file_name,
            ]
    # Only append the execution_uuid, tensorboard path and jobs flag if a real value was supplied
    if execution_uuid:
        cli_args.extend(["-e", execution_uuid])
    if tensorboard_path:
        cli_args.extend(["-t", tensorboard_path])
    if jobs:
        cli_args.extend(["-j", jobs])

    cli_args = cli.parse_args(cli_args)
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg

@exception_handler_decorator
def _repo_pull(pipeline_name: str, file_name: str, execution_uuid: str) -> str:
    """Pull artifacts, metadata files, and source code from the user's artifact
    repository, cmf-server, and git respectively.
    Args:
       pipeline_name: Name of the pipeline.
       file_name: Specify output metadata file name.
       execution_uuid: Specify execution uuid.
    Returns:
       Output from the repo pull command.
    """
    cli_args = cli.parse_args(
        [
            "repo",
            "pull",
            "-p",
            pipeline_name,
            "-f",
            file_name,
            "-e",
            execution_uuid,
        ]
    )
    cmd = cli_args.func(cli_args)
    msg = cmd.do_run()
    print(msg)
    return msg


@exception_handler_decorator
def _dvc_ingest(file_name: str) -> str:
   """Ingests metadata from the dvc.lock file into the CMF.
       If an existing MLMD file is provided, it merges and updates execution
       metadata based on matching commands, or creates new executions if none exist.
    Args:
       file_name: Specify input metadata file name.
    Returns:
       Output from the dvc ingest command.
   """
   cli_args = cli.parse_args(["dvc", "ingest", "-f", file_name])
   cmd = cli_args.func(cli_args)
   msg = cmd.do_run()
   print(msg)
   return msg



import os
from dotenv import load_dotenv
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, async_sessionmaker
from sqlalchemy.pool import NullPool  # For connection pooling (optional)
from server.app.db.dbmodels import metadata

# Load .env variables
load_dotenv()

# Database configuration
DB_CONFIG = {
    "db_user": os.getenv("POSTGRES_USER"),
    "db_password": os.getenv("POSTGRES_PASSWORD"),
    "db": os.getenv("POSTGRES_DB"),
    "db_host": os.getenv("POSTGRES_HOST"),
    "db_port": os.getenv("POSTGRES_PORT"),
}

#print(DB_CONFIG)


DATABASE_URL = "postgresql+asyncpg://{0}:{1}@{2}:{3}/{4}".format(
        DB_CONFIG["db_user"], DB_CONFIG["db_password"], DB_CONFIG["db_host"],
        DB_CONFIG["db_port"], DB_CONFIG["db"]
        )

#print("DATABASE_URL = ", DATABASE_URL)

# Create an asynchronous engine with connection pooling
engine = create_async_engine(
    DATABASE_URL,
    echo=True,  # For debugging, log SQL statements
)

# Create a session maker
async_session = async_sessionmaker(
    bind=engine,
    class_=AsyncSession,
    expire_on_commit=False,
)

async def get_db():
    async with async_session() as session:
        yield session


# Initialize DB schema (if not exists)
async def init_db():
    async with engine.begin() as conn:
        await conn.run_sync(metadata.create_all)



import pandas as pd
from cmflib.cmfquery import CmfQuery
import warnings
import typing as t

warnings.filterwarnings("ignore")

""" Old Code - not used anymore - keeping for future reference """

# def truncate_artifact_name(my_str):
#     if "/" in my_str:
#         my_str = my_str.split("/")[-1]
#     temp = my_str.split(":")
#     temp[1] = temp[1][:4]
#     temp=":".join(temp)
#     return temp

# def query_artifact_lineage_d3force(query: CmfQuery, pipeline_name, dict_of_art_ids: t.Dict[str, t.Dict[str, pd.DataFrame]]) -> dict:
#     art_name_id = {}
#     artifact_name_list = []
#     node_id_name_list = []
#     pipeline_id=query.get_pipeline_id(pipeline_name)
#     for type_, df in dict_of_art_ids[pipeline_name].items():
#         for index, row in df.iterrows():
#             artifact_name_list.append(row["name"])
#             art_name_id[row["name"]] = row["id"]
#             result="#48D1CC" if type_ == "Dataset" else "#2E8B57" if type_ =="Model" else "#FF8C00"
#             node_id_name_list.append({"name": truncate_artifact_name(row["name"]), "id": row["id"], "color": result})
#     link_list = []
#     for art_name in artifact_name_list:
#         immediate_child_artifacts = query.get_one_hop_child_artifacts(art_name, pipeline_id)
#         if immediate_child_artifacts.empty == True:
#             pass
#         else:
#             for i in list(set(immediate_child_artifacts["name"])):
#                 link = {}
#                 srce_art_id = art_name_id[art_name]
#                 trgt_art_id = art_name_id[i]
#                 link["source"] = int(srce_art_id)
#                 link["target"] = (trgt_art_id)
#                 link_list.append(link)

#     new_list = pd.DataFrame(
#     link_list
#     ).drop_duplicates().to_dict('records')
#     data = {
#         "nodes" : node_id_name_list,
#         "links" : new_list
#     }
#     return data

#print(query_visualization("/home/chobey/repair_lineage/testenv/example-get-started/mlmd","Test-env"))



###
# Copyright (2025) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import yaml
import uuid
import argparse
import typing as t
import pandas as pd

from cmflib.cmfquery import CmfQuery
from cmflib.cmf import Cmf
from cmflib.cli.command import CmdBase
from cmflib.dvc_wrapper import git_get_repo
from ml_metadata.metadata_store import metadata_store
from ml_metadata.proto import metadata_store_pb2 as mlpb
from cmflib.utils.helper_functions import fetch_cmf_config_path
from cmflib.cmf_exception_handling import (
    MsgSuccess, 
    FileNotFound, 
    DuplicateArgumentNotAllowed, 
    MissingArgument
)

class CmdDVCIngest(CmdBase):
    
    def find_location(self, search_element: str, list_of_elements: t.List[str]) -> t.Optional[int]:
        for index, element in enumerate(list_of_elements):
            if search_element == element:
                return index
        return None

    def ingest_metadata(self, *args) -> t.Dict[str, bool]:
        """
        Ingest the metadata into cmf
        
        args:
            A tuple of arguments in the following order:
            
            1. execution (str): The execution identifier or filename. 
            - Example: "3" (if already exists in CMF) or "demo_train.py" (if creating new execution).
            2. context_name (str): Context in which this execution is run. 
            - Example: "eval", "active_learning"
            3. pipeline_name (str): Name of the pipeline associated with the execution.
            - Example: "active_learning"
            4. metadata (dict): Parsed metadata dictionary from `dvc.lock` file.
            - Contains "deps" and "outs" describing inputs and outputs with MD5s.
            5. metawriter (object): CMF metadata writer object used to log executions, datasets, etc.
            6. tracked (dict): Dictionary tracking existing datasets that are already logged.
            - Key: dataset path
            - Value: True
            7. command (str): The command that was run for this execution.

        Returns:
            tracked (dict): Updated tracked dictionary after logging input/output datasets.
        """
        execution, context_name, pipeline_name, metadata, metawriter, tracked, command = args

        _ = metawriter.create_execution(
            str(context_name) + '_' + str(execution), 
            {}, 
            cmd = str(command),
            create_new_execution = False
            )
            
        output, config_file_path = fetch_cmf_config_path()
        artifact_repo = list(output.values())[0]

        for options, data in metadata.items():
            props = {}
            props["git_repo"] = git_get_repo()            
            if options == "deps":
                for dep in data:
                    # A dependency path (dep["path"]) can be a full file path like "artifacts/parsed/train.tsv"
                    # or just a directory like "artifacts/parsed".
                    if dep["path"] not in tracked:
                        metawriter.log_dataset(dep["path"], 'input')
                    else:
                        md5_value = dep["md5"] 
                        url = pipeline_name + ":" + artifact_repo + "/files/md5/" + md5_value[:2] + "/" + md5_value[2:]
                        props["url"] = url
                        metawriter.log_dataset_with_version(dep["path"], md5_value, "input", props)
            if options == "outs":
                for out in data:
                    md5_value = out["md5"] 
                    url = pipeline_name + ":" + artifact_repo + "/files/md5/" + md5_value[:2] + "/" + md5_value[2:]
                    props["url"] = url
                    metawriter.log_dataset_with_version(out["path"], md5_value, "output", props)
                    tracked[out["path"]] = True
        return tracked

    def run(self, live):
        """
            Ingest the metadata in dvc.lock file into CMF.
            If mlmd file exist with metadata from metrics logging or logging for other
            metadata not captured in dvc.lock file, pass that mlmd file as the input file.
            This code queries the file for existing pipelines, stages and executions and stores 
            as a dictionary. If the execution cmd in the stored dict, matches the execution command
            in the dvc.lock file, that execution is updated with additional metadata.
            If there is no prior execution captured, a new execution is created
        """
        cmd_args = {
            "file_name": self.args.file_name,
        }
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
                
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
         
        current_directory = os.path.dirname(mlmd_file_name)
        # In our case, "mlmd" is the default file name. Whether the file exists or not, we need to continue with the same name.
        if not os.path.exists(mlmd_file_name) and mlmd_file_name != "./mlmd":  
            raise FileNotFound(mlmd_file_name, current_directory)

        uuid_ = str(uuid.uuid4())
        pipeline_name = ""
        cmd_exe: t.Dict[str, str] = {}
        query = cmfquery.CmfQuery(mlmd_file_name)
        pipelines: t.List[str] = query.get_pipeline_names()

        # Query mlmd to get all the executions and its commands
        for pipeline in pipelines:
            pipeline_name = pipeline
            stages: t.List[str] = query.get_pipeline_stages(pipeline)
            for stage in stages:
                exe_df: pd.DataFrame = query.get_all_executions_in_stage(stage)
                """
                Parse all the executions in a stage
                eg- exe_step = ['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']
                """
                for index, row in exe_df.iterrows():
                    exe_step = row['Execution']
                    '''
                    if already same execution command has been captured previously use the latest
                    execution id to associate the new metadata
                    '''
                    existing = cmd_exe.get(exe_step)
                    if existing is None:
                        cmd_exe[exe_step] = f"{row['id']},{stage},{pipeline}"
                    else:
                        if row['id'] > int(existing.split(',')[0]):
                            cmd_exe[exe_step] = f"{row['id']},{stage},{pipeline}"

        # Parse the dvc.lock file.
        # pipeline_dict stores pipeline stages with the following hierarchy:
        # {
        #     "<stage_name>": {
        #         "<index>": {
        #             "cmd": List[str],        # List of command parts as strings
        #             "deps": List[Dict[str, Any]],  # List of dependency dictionaries
        #             "outs": List[Dict[str, Any]]   # List of output dictionaries
        #         }
        #     }
        # }
        pipeline_dict: t.Dict[str, t.Dict[str, t.Dict[str, t.Union[t.List[str], t.List[t.Dict[str, t.Any]]]]]] = {}
        
        # check whether dvc.lock file present in cwd
        if not os.path.exists("dvc.lock"):
            raise FileNotFound("dvc.lock", os.getcwd())

        with open("dvc.lock", 'r') as f:
            valuesYaml = yaml.load(f, Loader=yaml.FullLoader)

        for stage in valuesYaml['stages']:
            pipeline_dict[stage] = {}
            commands=[]
            deps = []
            outs = []
            k_dict = {}
            i = 0
            
            for options in valuesYaml['stages'][stage]:
                if options == 'cmd':
                    cmd_list = valuesYaml['stages'][stage][options].split()
                    commands.append(cmd_list)
                    k_dict['cmd'] = cmd_list
                    i = i + 1
                if options == 'deps':
                    deps = valuesYaml['stages'][stage][options]
                    k_dict['deps'] = deps
                if options == 'outs':
                    outs = valuesYaml['stages'][stage][options]
                    k_dict['outs'] = outs

            pipeline_dict[stage][str(i)] = k_dict

        # Create a unique Pipeline name if there is no mlmd file
        pipeline_name = "Pipeline" + "-" + str(uuid_) if not pipeline_name else pipeline_name
        metawriter = Cmf(filepath = mlmd_file_name, pipeline_name = pipeline_name, graph = True)

        # Parse the dvc.lock dictionary and get the command section
        tracked = {} #Used to keep a record of files tracked by outs and therefore not needed to be tracked in deps
        for stage, stage_data in pipeline_dict.items():
            for index, dict_data in stage_data.items():
                for options, option_details in dict_data.items():
                    if options == 'cmd':
                        """
                        Key eg - cmd
                        Value eg - ['python3', 'demo.py', '--enable_df', 'True']
                        cmd_exe eg - {"['demo_eval.py', '--trained_model', 'data/model-3', '--enable_df', 'True', '--round', '3']":
                        '3,eval,active_learning', 
                        "['demo_eval.py', '--trained_model', 'data/model-2', '--enable_df', 'True', '--round', '2']": '2,eval,active_learning',
                        "['demo_eval.py', '--trained_model', 'data/model-1', '--enable_df', 'True', '--round', '1']": '1,eval,active_learning'}
                        In the next line pop out the python
                        if the pipeline_dict command is already there in the cmd_exe dict got from parsing the mlmd pop that cmd out 
                        and use the stored lineage from the mlmd
                        """
                        option_details.pop(0)
                        # This check is specific to the active learning scenario where '--execution_name' is expected.
                        pos = self.find_location('--execution_name', option_details)
                        if pos:
                            execution_name = option_details[pos+1]
                        else:
                            execution_name = uuid_
                            
                        context_name = stage
                        # Cast option_details to a list of strings to ensure join works correctly.
                        # join the list to single string
                        cmd_str = ' '.join(t.cast(t.List[str], option_details))
                        cmd = cmd_exe.get(cmd_str, None)
                        _ = metawriter.create_context(pipeline_stage=context_name)
                        tracked = self.ingest_metadata(execution_name, context_name, pipeline_name, dict_data, metawriter, tracked, cmd_str)
        metawriter.log_dvc_lock("dvc.lock")
        return MsgSuccess(msg_str="Done.")


def add_parser(subparsers, parent_parser):
    HELP = "Ingests metadata from the dvc.lock file into the CMF."

    parser = subparsers.add_parser(
        "ingest",
        parents=[parent_parser],
        description="Ingests metadata from the dvc.lock file into the CMF. If an existing metadata file is provided, it merges and updates execution metadata based on matching commands; otherwise, it creates new executions.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input mlmd file name. (default: mlmd)",
        metavar="<file_name>"
    )

    parser.set_defaults(func=CmdDVCIngest)
